import sys

PY2 = sys.version_info < (3,)


def download(url, dest=None):
    import os

    if PY2:
        from urllib import urlretrieve
    else:
        from urllib.request import urlretrieve

    if dest is None:
        dest = os.getcwd()

    filepath = os.path.join(dest, _filename(url))
    urlretrieve(url, filepath)


def _filename(url):
    import os

    if PY2:
        from urlparse import urlparse
    else:
        from urllib.parse import urlparse

    a = urlparse(url)
    return os.path.basename(a.path)



from numpy.testing import assert_allclose


def _generate_data(N, P, K, S):
    import scipy as sp

    # fixed eff
    F = sp.randn(N, K)
    B0 = 2 * (sp.arange(0, K * P) % 2) - 1.0
    B0 = sp.reshape(B0, (K, P))
    FB = sp.dot(F, B0)

    # gerenrate phenos
    h2 = sp.linspace(0.1, 0.5, P)

    # generate data
    G = 1.0 * (sp.rand(N, S) < 0.2)
    G -= G.mean(0)
    G /= G.std(0)
    G /= sp.sqrt(G.shape[1])
    Wg = sp.randn(P, P)
    Wn = sp.randn(P, P)

    B = sp.randn(G.shape[1], Wg.shape[1])
    Yg = G.dot(B).dot(Wg.T)
    Yg -= Yg.mean(0)
    Yg *= sp.sqrt(h2 / Yg.var(0))
    Cg0 = sp.cov(Yg.T)

    B = sp.randn(G.shape[0], Wg.shape[1])
    Yn = B.dot(Wn.T)
    Yn -= Yn.mean(0)
    Yn *= sp.sqrt((1 - h2) / Yn.var(0))
    Cn0 = sp.cov(Yn.T)

    Y = FB + Yg + Yn

    return Y, F, G, B0, Cg0, Cn0


def test_mtmm_scan_pv_beta():
    import scipy as sp
    import scipy.linalg as la
    from limix_core.gp import GP2KronSum
    from limix_core.covar import FreeFormCov

    N = 200
    P = 4
    M = 2
    K = 2
    S = 10
    Y, F, G, B0, Cg0, Cn0 = _generate_data(N, P, K, S)
    A = sp.eye(P)
    Asnp = sp.rand(P, M)

    # compute eigenvalue decomp of RRM
    R = sp.dot(G, G.T)
    R /= R.diagonal().mean()
    R += 1e-4 * sp.eye(R.shape[0])
    Sr, Ur = la.eigh(R)

    # fit null model
    Cg = FreeFormCov(Y.shape[1])
    Cn = FreeFormCov(Y.shape[1])
    gp = GP2KronSum(Y=Y, S_R=Sr, U_R=Ur, Cg=Cg, Cn=Cn, F=F, A=sp.eye(P))
    gp.covar.Cg.setCovariance(0.5 * sp.cov(Y.T))
    gp.covar.Cn.setCovariance(0.5 * sp.cov(Y.T))
    gp.optimize(factr=10)

    # run MTLMM
    from limix_lmm import MTLMM

    mtlmm = MTLMM(Y, F=F, A=A, Asnp=Asnp, covar=gp.covar)
    pv, B = mtlmm.process(G)

    # run standard LMMcore
    from limix_lmm.lmm_core import LMMCore

    y = sp.reshape(Y, [Y.size, 1], order="F")
    covs = sp.kron(A, F)
    Aext = sp.kron(Asnp, sp.ones((G.shape[0], 1)))
    Gext = sp.kron(sp.ones((Asnp.shape[0], 1)), G)
    Wext = sp.einsum("ip,in->inp", Aext, Gext).reshape(Aext.shape[0], -1)
    stlmm = LMMCore(y, covs, Ki_dot=gp.covar.solve)
    stlmm.process(Wext, step=Asnp.shape[1])
    pv0 = stlmm.getPv()
    B0 = stlmm.getBetaSNP()

    assert_allclose(pv0, pv, rtol=1e-06, atol=1e-06)
    assert_allclose(B0, B, rtol=1e-06, atol=1e-06)



import os
import numpy as np
import pandas as pd
import scipy as sp
import scipy.linalg as la
from limix_core.util.preprocess import gaussianize
from limix_lmm import GWAS_LMM, GWAS_MTLMM
from limix_lmm import download, unzip
from pandas_plink import read_plink
from sklearn.impute import SimpleImputer
import geno_sugar as gs
import geno_sugar.preprocess as prep


if __name__ == "__main__":

    # download data
    download("http://www.ebi.ac.uk/~casale/data_structlmm.zip")
    unzip("data_structlmm.zip")

    # import snp data
    bedfile = "data_structlmm/chrom22_subsample20_maf0.10"
    (bim, fam, G) = read_plink(bedfile, verbose=False)

    # consider the first 100 snps
    snps = G[:100].compute().T

    # define genetic relatedness matrix
    W_R = sp.randn(fam.shape[0], 20)
    R = sp.dot(W_R, W_R.T)
    R /= R.diagonal().mean()
    S_R, U_R = la.eigh(R)

    # load phenotype data
    phenofile = "data_structlmm/expr.csv"
    dfp = pd.read_csv(phenofile, index_col=0)
    pheno = gaussianize(dfp.loc["gene1"].values[:, None])

    # define covs
    covs = sp.ones([pheno.shape[0], 1])

    # linear model test
    lm = GWAS_LMM(pheno, covs=covs, verbose=True)
    res = lm.process(snps)
    print(res.head())

    # linear mixed model
    lmm = GWAS_LMM(pheno, covs=covs, eigh_R=(S_R, U_R), verbose=True)
    res = lmm.process(snps)
    print(res.head())

    # low-rank linear mixed model (low-rank)
    lrlmm = GWAS_LMM(pheno, covs=covs, W_R=W_R, verbose=True)
    res = lrlmm.process(snps)
    print(res.head())

    # generate interacting variables to test
    inter = sp.randn(pheno.shape[0], 1)

    # interaction test
    lmi = GWAS_LMM(pheno, covs=covs, inter=inter, verbose=True)
    res = lmi.process(snps)
    print(res.head())

    # generate interacting variables to test
    inter0 = sp.randn(pheno.shape[0], 1)

    # interaction test
    lmi = GWAS_LMM(pheno, covs=covs, inter=inter, inter0=inter0, verbose=True)
    res = lmi.process(snps)
    print(res.head())

    # test multi-trait association
    P = 4
    phenos = sp.randn(pheno.shape[0], P)
    Asnps = sp.eye(P)
    mtlmm = GWAS_MTLMM(phenos, covs=covs, Asnps=Asnps, eigh_R=(S_R, U_R), verbose=True)
    res = mtlmm.process(snps)
    print(res.head())

    # common effect test
    Asnps0 = sp.ones([P, 1])
    mtlmm = GWAS_MTLMM(
        phenos, covs=covs, Asnps=Asnps, Asnps0=Asnps0, eigh_R=(S_R, U_R), verbose=True
    )
    res = mtlmm.process(snps)
    print(res.head())

    # slice of genome to analyze
    Isnp = gs.is_in(bim, ("22", 17500000, 18000000))
    G, bim = gs.snp_query(G, bim, Isnp)

    # define geno preprocessing function for geno-wide analysis
    imputer = SimpleImputer(missing_values=np.nan, strategy="mean")
    preprocess = prep.compose(
        [
            prep.filter_by_missing(max_miss=0.10),
            prep.impute(imputer),
            prep.filter_by_maf(min_maf=0.10),
            prep.standardize(),
        ]
    )

    from limix_lmm.util import append_res

    # slide large genetic region using batches of 200 variants
    res = []
    queue = gs.GenoQueue(G, bim, batch_size=200, preprocess=preprocess)
    for _G, _bim in queue:

        _res = {}
        _res["lm"] = lm.process(_G)
        _res["lmm"] = lmm.process(_G)
        _res["lrlmm"] = lrlmm.process(_G)
        _res = append_res(_bim, _res)
        _res.append(_res)

    # export
    print("Exporting to out/")
    if not os.path.exists("out"):
        os.makedirs("out")
    res = pd.concat(res)
    res.reset_index(inplace=True, drop=True)
    res.to_csv("out/res_lmm.csv", index=False)



import scipy as sp

from .util import christof_trick


def define_helper(Y, F, A, Asnp, covar=None):
    import limix_core

    if covar is None:
        return MTLMMHelperBase(Y, F, A, Asnp)
    else:
        # at the moment only Cov2KronSum is supported but
        # more covariances can be added within the same framework
        covtypes = [limix_core.covar.Cov2KronSum]
        assert type(covar) in covtypes, "This covariance type is not supported"
        if type(covar == limix_core.covar.Cov2KronSum):
            return MTLMMHelper2KronSum(Y, F, A, Asnp, covar)


class MTLMMHelperBase:
    r"""
    Internal helper class for MTLMM
    """

    def __init__(self, Y, F, A, Asnp):
        self._Y = Y
        self._F = F
        self._A = A
        self._Asnp = Asnp
        self._YAsnp = sp.dot(self._Y, self._Asnp)

    def set_g(self, _g):
        self._g = _g

    def get_WKiy(self):
        RV = sp.dot(self._F.T, self._Y.dot(self._A))
        RV = sp.reshape(RV, [RV.size, 1], order="F")
        return RV

    def get_yKiy(self):
        return sp.einsum("ip,ip->", self.LY, self._Y)

    def get_WKiW(self):
        return christof_trick(self._A, self._F, sp.ones(self.Y.size))

    def get_W1Kiy_2(self):
        return sp.dot(self._YAsnp.T, self._g)

    def get_W1KiW1_12(self):
        return christof_trick(
            self._A, self._F, sp.ones(self.Y.size), A2=self._Asnp, F2=self._g
        )

    def get_W1KiW1_22(self):
        return christof_trick(self._Asnp, self._g, sp.ones(self.Y.size))


class MTLMMHelper2KronSum(MTLMMHelperBase):
    r"""
    Internal helper class for MTLMM
    """

    def __init__(self, Y, F, A, Asnp, covar):
        self.Y = Y
        self.F = F
        self.A = A
        self.Asnp = Asnp
        self.covar = covar

        # compute stuff from covariance
        self.LY = covar.Lr().dot(self.Y.dot(covar.Lc().T))
        self._Y = covar.D() * self.LY
        self._A = covar.Lc().dot(self.A)
        self._Asnp = covar.Lc().dot(self.Asnp)
        self._F = covar.Lr().dot(self.F)
        self._D = self.covar.D()
        self._YAsnp = sp.dot(self._Y, self._Asnp)

    def set_g(self, g):
        self._g = self.covar.Lr().dot(g)

    def get_WKiy(self):
        RV = sp.dot(self._F.T, self._Y.dot(self._A))
        RV = sp.reshape(RV, [RV.size, 1], order="F")
        return RV

    def get_yKiy(self):
        return sp.einsum("ip,ip->", self.LY, self._Y)

    def get_WKiW(self):
        return christof_trick(self._A, self._F, self.covar.D())

    def get_W1Kiy_2(self):
        return sp.dot(self._YAsnp.T, self._g)

    def get_W1KiW1_12(self):
        return christof_trick(
            self._A, self._F, self.covar.D(), A2=self._Asnp, F2=self._g
        )

    def get_W1KiW1_22(self):
        return christof_trick(self._Asnp, self._g, self.covar.D())



from time import time

from .lmm_core import LMMCore


class LMM(LMMCore):
    r"""
    Standard LMM with general bg covariance

    The LMM model is

    .. math::
        \mathbf{y}=\sim\mathcal{N}(
        \underbrace{\mathbf{F}\mathbf{b}}_{\text{covariates}}+
        \underbrace{\mathbf{x}\beta}_{\text{genetics}},
        \underbrace{\mathbf{K}_{\boldsymbol{\theta}}}_{\text{covariance}})

    The test :math:`\beta\neq{0}` is done for all provided variants
    one-by-one.

    Parameters
    ----------
    y : (`N`, 1) ndarray
        phenotype vector
    F : (`N`, L) ndarray
        fixed effect design for covariates.
    Ki_dot : function
        method that takes an array and returns the dot product of
        the inverse of the covariance and the input array.

    Examples
    --------

    .. doctest::

        >>> from numpy.random import RandomState
        >>> import scipy as sp
        >>> from limix_lmm import LMM
        >>> from limix_core.gp import GP2KronSumLR
        >>> from limix_core.covar import FreeFormCov
        >>> random = RandomState(1)
        >>> from numpy import set_printoptions
        >>> set_printoptions(4)
        >>>
        >>> # generate data
        >>> N = 100
        >>> k = 1
        >>> S = 1000
        >>> y = random.randn(N, 1)
        >>> E = random.randn(N, k)
        >>> G = 1.*(random.rand(N, S) < 0.2)
        >>> F = random.rand(N, 1)
        >>> F = sp.concatenate([sp.ones((N, 1)), F], 1)
        >>>
        >>> # larn a covariance on the null model
        >>> gp = GP2KronSumLR(Y=y, Cn=FreeFormCov(1), G=E, F=F, A=sp.ones((1,1)))
        >>> gp.covar.Cr.setCovariance(0.5*sp.ones((1,1)))
        >>> gp.covar.Cn.setCovariance(0.5*sp.ones((1,1)))
        >>> info_opt = gp.optimize(verbose=False)
        >>>
        >>> lmm = LMM(y, F, gp.covar.solve)
        >>> lmm.process(G)
        >>> pv = lmm.getPv()
        >>> beta = lmm.getBetaSNP()
        >>> beta_ste = lmm.getBetaSNPste()
        >>> lrt = lmm.getLRT()
        >>>
        >>> print(pv[:4])
        [0.8335 0.1669 0.9179 0.279 ]
        >>> print(beta[:4])
        [-0.0479  0.3145  0.0235 -0.2465]
        >>> print(beta_ste[:4])
        [0.2279 0.2275 0.2283 0.2276]
        >>> print(lrt[:4])
        [0.0442 1.9108 0.0106 1.1721]
    """

    def __init__(self, y, F, Ki_dot=None):
        import scipy as sp

        if F is None:
            F = sp.ones((y.shape[0], 1))
        self.y = y
        self.F = F
        self.Ki_dot = Ki_dot
        self.df = y.shape[0] - F.shape[1]
        self._fit_null()

    def _fit_null(self):
        """ Internal functon. Fits the null model """
        import scipy as sp
        import scipy.linalg as la

        if self.Ki_dot is None:
            self.Kiy = self.y
            self.KiF = self.F
        else:
            self.Kiy = self.Ki_dot(self.y)
            self.KiF = self.Ki_dot(self.F)
        self.FKiy = sp.dot(self.F.T, self.Kiy)
        self.FKiF = sp.dot(self.F.T, self.KiF)
        self.yKiy = sp.dot(self.y[:, 0], self.Kiy[:, 0])
        # calc beta_F0 and s20
        self.A0i = la.inv(self.FKiF)
        self.beta_F0 = sp.dot(self.A0i, self.FKiy)
        self.s20 = (self.yKiy - sp.dot(self.FKiy[:, 0], self.beta_F0[:, 0])) / self.df

    def process(self, G, verbose=False):
        r"""
        Fit genotypes one-by-one.

        Parameters
        ----------
        G : (`N`, `S`) ndarray
            genotype vector for `N` individuals and `S` variants.
        verbose : bool
            verbose flag.

        Returns
        -------
        pv : ndarray
            P values
        beta : ndarray
            variant effect szies
        """
        import scipy as sp
        import scipy.stats as st

        t0 = time()
        # precompute some stuff
        if self.Ki_dot is None:
            KiG = G
        else:
            KiG = self.Ki_dot(G)
        GKiy = sp.dot(G.T, self.Kiy[:, 0])
        GKiG = sp.einsum("ij,ij->j", G, KiG)
        FKiG = sp.dot(self.F.T, KiG)

        # Let us denote the inverse of Areml as
        # Ainv = [[A0i + m mt / n, m], [mT, n]]
        A0iFKiG = sp.dot(self.A0i, FKiG)
        n = 1.0 / (GKiG - sp.einsum("ij,ij->j", FKiG, A0iFKiG))
        M = -n * A0iFKiG
        self.beta_F = self.beta_F0 + M * sp.dot(M.T, self.FKiy[:, 0]) / n
        self.beta_F += M * GKiy
        self.beta_g = sp.einsum("is,i->s", M, self.FKiy[:, 0])
        self.beta_g += n * GKiy

        # sigma
        s2 = self.yKiy - sp.einsum("i,is->s", self.FKiy[:, 0], self.beta_F)
        s2 -= GKiy * self.beta_g
        s2 /= self.df

        # dlml and pvs
        self.lrt = -self.df * sp.log(s2 / self.s20)
        self.pv = st.chi2(1).sf(self.lrt)

        t1 = time()
        if verbose:
            print("Tested for %d variants in %.2f s" % (G.shape[1], t1 - t0))



from time import time

from .util import calc_Ai_beta_s2, hatodot


class LMMCore:
    r"""
    Core LMM for interaction and association testing.

    The model is

    .. math::
        \mathbf{y}\sim\mathcal{N}(
        \underbrace{\mathbf{F}\mathbf{b}}_{\text{covariates}}+
        \underbrace{(\mathbf{X}\hat{\odot}\mathbf{I})\;
        \boldsymbol{\beta}}_{\text{genetics}},
        \sigma^2\underbrace{\mathbf{K}_{
        \boldsymbol{\theta}_0}}_{\text{covariance}})

    :math:`\hat{\odot}` is defined as follows.
    Let

    .. math::
        \mathbf{A}=\left[\mathbf{a}_1,
                         \dots,
                         \mathbf{a}_{m}
                         \right]\in\mathbb{R}^{n\times{m}}]

    and

    .. math::
        \mathbf{B}=\left[\mathbf{b}_1,
                         \dots,
                         \mathbf{b}_{l}
                         \right]\in\mathbb{R}^{n\times{l}}

    then

    .. math::
        \mathbf{A}\hat{\odot}\mathbf{B} =
         \left[\mathbf{a}_1\odot\mathbf{b}_1
               \dots,
               \mathbf{a}_1\odot\mathbf{b}_{l},
               \mathbf{a}_2\odot\mathbf{b}_1,
               \dots,
               \mathbf{a}_2\odot\mathbf{b}_{l},
               \dots,
               \mathbf{a}_{m}\odot\mathbf{b}_{l},
               \dots,\mathbf{a}_{m}\odot\mathbf{b}_{l}
               \right]\in\mathbb{R}^{n\times{(ml)}}

    where :math:`\odot` is the element-wise product.

    Importantly, :math:`\mathbf{K}_{\boldsymbol{\theta}_0}`
    is not re-estimated under the laternative model,
    i.e. only :math:`\sigma^2` is learnt.
    The covariance parameters :math:`\boldsymbol{\theta}_0`
    need to be learnt using another module, e.g. limix-core.
    Note that as a consequence of this,
    such an implementation of single-variant analysis does not require the
    specification of the whole covariance but just of a method that
    specifies the product of its inverse by a vector.

    The test :math:`\boldsymbol{\beta}\neq{0}` is done
    in bocks of ``step`` variants,
    where ``step`` can be specifed by the user.

    Parameters
    ----------
    y : (`N`, 1) ndarray
        phenotype vector
    F : (`N`, L) ndarray
        fixed effect design for covariates.
    Ki_dot : function
        method that takes an array and returns the dot product of
        the inverse of the covariance and the input array.


    Examples
    --------

        Example with Inter and step=1

        .. doctest::

            >>> from numpy.random import RandomState
            >>> import scipy as sp
            >>> from limix_lmm.lmm_core import LMMCore
            >>> from limix_core.gp import GP2KronSumLR
            >>> from limix_core.covar import FreeFormCov
            >>> random = RandomState(1)
            >>> from numpy import set_printoptions
            >>> set_printoptions(4)
            >>>
            >>> N = 100
            >>> k = 1
            >>> m = 2
            >>> S = 1000
            >>> y = random.randn(N,1)
            >>> E = random.randn(N,k)
            >>> G = 1.*(random.rand(N,S)<0.2)
            >>> F = sp.concatenate([sp.ones((N,1)), random.randn(N,1)], 1)
            >>> Inter = random.randn(N, m)
            >>>
            >>> gp = GP2KronSumLR(Y=y, Cn=FreeFormCov(1), G=E, F=F, A=sp.ones((1,1)))
            >>> gp.covar.Cr.setCovariance(0.5*sp.ones((1,1)))
            >>> gp.covar.Cn.setCovariance(0.5*sp.ones((1,1)))
            >>> info_null = gp.optimize(verbose=False)
            >>>
            >>> lmm = LMMCore(y, F, gp.covar.solve)
            >>> lmm.process(G, Inter)
            >>> pv = lmm.getPv()
            >>> beta = lmm.getBetaSNP()
            >>> beta_ste = lmm.getBetaSNPste()
            >>> lrt = lmm.getLRT()
            >>>
            >>> print(pv.shape)
            (1000,)
            >>> print(beta.shape)
            (2, 1000)
            >>> print(pv[:4])
            [0.1428 0.3932 0.4749 0.3121]
            >>> print(beta[:,:4])
            [[ 0.0535  0.2571  0.122  -0.2328]
             [ 0.3418  0.3179 -0.2099  0.0986]]

        Example with step=4

        .. doctest::

            >>> lmm.process(G, step=4)
            >>> pv = lmm.getPv()
            >>> beta = lmm.getBetaSNP()
            >>> lrt = lmm.getLRT()
            >>>
            >>> print(pv.shape)
            (250,)
            >>> print(beta.shape)
            (4, 250)
            >>> print(pv[:4])
            [0.636  0.3735 0.7569 0.3282]
            >>> print(beta[:,:4])
            [[-0.0176 -0.4057  0.0925 -0.4175]
             [ 0.2821  0.2232 -0.2124  0.2433]
             [-0.0575  0.1528 -0.0384  0.019 ]
             [-0.2156 -0.2327 -0.1773 -0.1497]]

        Example with step=4 and Inter

        .. doctest::

            >>> lmm = LMMCore(y, F, gp.covar.solve)
            >>> lmm.process(G, Inter=Inter, step=4)
            >>> pv = lmm.getPv()
            >>> beta = lmm.getBetaSNP()
            >>> lrt = lmm.getLRT()
            >>>
            >>> print(pv.shape)
            (250,)
            >>> print(beta.shape)
            (8, 250)
            >>> print(pv[:4])
            [0.1591 0.2488 0.4109 0.5877]
            >>> print(beta[:,:4])
            [[ 0.0691  0.1977  0.435  -0.3214]
             [ 0.1701 -0.3195  0.0179  0.3344]
             [ 0.1887 -0.0765 -0.3847  0.1843]
             [-0.2974  0.2787  0.2427 -0.0717]
             [ 0.3784  0.0854  0.1566  0.0652]
             [ 0.4012  0.5255 -0.1572  0.1674]
             [-0.413   0.0278  0.1946 -0.1199]
             [ 0.0268 -0.0317 -0.1059  0.1414]]
    """

    def __init__(self, y, F, Ki_dot=None):
        import scipy as sp

        if F is None:
            F = sp.ones((y.shape[0], 1))
        self.y = y
        self.F = F
        self.Ki_dot = Ki_dot
        self.df = y.shape[0] - F.shape[1]
        self._fit_null()

    def _fit_null(self):
        """ Internal functon. Fits the null model """
        import scipy as sp

        if self.Ki_dot is None:
            self.Kiy = self.y
            self.KiF = self.F
        else:
            self.Kiy = self.Ki_dot(self.y)
            self.KiF = self.Ki_dot(self.F)
        self.FKiy = sp.dot(self.F.T, self.Kiy)
        self.FKiF = sp.dot(self.F.T, self.KiF)
        self.yKiy = sp.dot(self.y[:, 0], self.Kiy[:, 0])
        # calc beta_F0 and s20
        self.A0i, self.beta_F0, self.s20 = calc_Ai_beta_s2(
            self.yKiy, self.FKiF, self.FKiy, self.df
        )

    def process(self, G, Inter=None, step=1, verbose=False):
        r"""
        Fit genotypes one-by-one.

        Parameters
        ----------
        G : (`N`, `S`) ndarray
        Inter : (`N`, `M`) ndarray
            Matrix of `M` factors for `N` inds with which
            each variant interact
            By default, Inter is set to a matrix of ones.
        step : int
            Number of consecutive variants that should be
            tested jointly.
        verbose : bool
            verbose flag.
        """
        import scipy as sp
        import scipy.stats as st

        t0 = time()
        ntests = int(G.shape[1] / step)
        if Inter is None:
            mi = 1
        else:
            mi = Inter.shape[1]
        k = self.F.shape[1]
        m = mi * step
        F1KiF1 = sp.zeros((k + m, k + m))
        F1KiF1[:k, :k] = self.FKiF
        F1Kiy = sp.zeros((k + m, 1))
        F1Kiy[:k, 0] = self.FKiy[:, 0]
        s2 = sp.zeros(ntests)
        self.beta_g = sp.zeros([m, ntests])
        for s in range(ntests):
            idx1 = step * s
            idx2 = step * (s + 1)
            if Inter is not None:
                if step == 1:
                    X = Inter * G[:, idx1:idx2]
                else:
                    X = hatodot(Inter, G[:, idx1:idx2])
            else:
                X = G[:, idx1:idx2]
            if self.Ki_dot is None:
                KiX = X
            else:
                KiX = self.Ki_dot(X)
            F1KiF1[k:, :k] = sp.dot(X.T, self.KiF)
            F1KiF1[:k, k:] = F1KiF1[k:, :k].T
            F1KiF1[k:, k:] = sp.dot(X.T, KiX)
            F1Kiy[k:, 0] = sp.dot(X.T, self.Kiy[:, 0])
            # this can be sped up by using block matrix inversion, etc
            _, beta, s2[s] = calc_Ai_beta_s2(self.yKiy, F1KiF1, F1Kiy, self.df)
            self.beta_g[:, s] = beta[k:, 0]
        # dlml and pvs
        self.lrt = -self.df * sp.log(s2 / self.s20)
        self.pv = st.chi2(m).sf(self.lrt)

        t1 = time()
        if verbose:
            print("Tested for %d variants in %.2f s" % (G.shape[1], t1 - t0))

    def getPv(self):
        """
        Get pvalues

        Returns
        -------
        pv : ndarray
        """
        return self.pv

    def getBetaSNP(self):
        """
        get effect size SNPs


        Returns
        -------
        pv : ndarray
        """
        return self.beta_g

    def getBetaCov(self):
        """
        get beta of covariates

        Returns
        -------
        beta : ndarray
        """
        return self.beta_F

    def getLRT(self):
        """
        get lik ratio test statistics

        Returns
        -------
        lrt : ndarray
        """
        return self.lrt

    def getBetaSNPste(self):
        """
        get standard errors on betas

        Returns
        -------
        beta_ste : ndarray
        """
        import scipy as sp
        import scipy.stats as st

        beta = self.getBetaSNP()
        pv = self.getPv()
        z = sp.sign(beta) * sp.sqrt(st.chi2(1).isf(pv))
        ste = beta / z
        return ste



r"""
*********
Limix-LMM
*********

Flexible Linear mixed model Toolbox for Genome-wide association studies

"""

from __future__ import absolute_import as _

from . import plot
from .lmm import LMM
from .lmm_core import LMMCore
from .mtlmm import MTLMM
from ._testit import test
from .sh import download, unzip


__version__ = "0.1.1"

__all__ = [
    "LMM",
    "LMMCore",
    "MTLMM",
    "plot",
    "download",
    "unzip",
    "__version__",
    "test",
]



def qqplot(
    ax,
    pv,
    color="k",
    plot_method=None,
    line=False,
    pv_thr=1e-2,
    plot_xyline=False,
    xy_labels=False,
    xyline_color="r",
):
    """
    Utility function to make manhattan plot

    Parameters
    ----------
    ax : pyplot plot
        subplot
    df : pandas.DataFrame
        pandas DataFrame with chrom, pos and pv
    color : color
        colors to use in the manhattan plot (default is black)
    plot_method : function
        function that takes x and y and plots a custom scatter plot
        The default value is None.
    pv_thr : float
        threshold on pvs
    plot_xyline : bool
        if True, the x=y line is plotteed. The default value is False.

    Examples
    --------

    .. doctest::

        >>> from limix_lmm.plot import qqplot
        >>> import scipy as sp
        >>> from matplotlib import pyplot as plt
        >>>
        >>> pv1 = sp.rand(10000)
        >>> pv2 = sp.rand(10000)
        >>> pv3 = sp.rand(10000)
        >>>
        >>> ax = plt.subplot(111)
        >>> qqplot(ax, pv1, color='C0')
        >>> qqplot(ax, pv2, color='C1')
        >>> qqplot(ax, pv3, color='C2', plot_xyline=True, xy_labels=True)
    """
    import scipy as sp

    pv1 = pv[pv < pv_thr]
    pvo = -sp.log10(sp.sort(pv1))
    pvt = -sp.log10(sp.linspace(0, pv_thr, pv1.shape[0] + 2)[1:-1])
    if plot_method is not None:
        plot_method(pvt, pvo)
    else:
        ax.plot(pvt, pvo, marker=".", color=color)

    if plot_xyline:
        xlim1, xlim2 = ax.get_xlim()
        ylim1, ylim2 = ax.get_ylim()
        xlim1 = -sp.log10(pv_thr)
        ylim1 = -sp.log10(pv_thr)
        ax.plot([xlim1, xlim2], [xlim1, xlim2], xyline_color)
        ax.set_xlim(xlim1, xlim2)
        ax.set_ylim(ylim1, ylim2)
    if xy_labels:
        ax.set_xlabel("Expected P values")
        ax.set_ylabel("Observed P values")



def plot_manhattan(ax, df, pv_thr=None, colors=None, offset=None, callback=None):
    """
    Utility function to make manhattan plot

    Parameters
    ----------
    ax : pyplot plot
        subplot
    df : pandas.DataFrame
        pandas DataFrame with chrom, pos and pv
    colors : list
        colors to use in the manhattan plot
    offset : float
        offset between in chromosome expressed as fraction of the
        length of the longest chromosome (default is 0.2)
    callback : function
        callback function that takes as input df

    Examples
    --------

    .. doctest::

        >>> from matplotlib import pyplot as plt
        >>> from limix_lmm.plot import plot_manhattan
        >>> import scipy as sp
        >>> import pandas as pd
        >>> n_chroms = 5
        >>> n_snps_per_chrom = 10000
        >>> chrom = sp.kron(sp.arange(1, n_chroms + 1), sp.ones(n_snps_per_chrom))
        >>> pos = sp.kron(sp.ones(n_chroms), sp.arange(n_snps_per_chrom))
        >>> pv = sp.rand(n_chroms * n_snps_per_chrom)
        >>> df = pd.DataFrame({'chrom': chrom, 'pos': pos, 'pv': pv})
        >>>
        >>> ax = plt.subplot(111)
        >>> plot_manhattan(ax, df)
    """
    import scipy as sp

    if colors is None:
        colors = ["k", "Gray"]
    if offset is None:
        offset = 0.2
    dx = offset * df["pos"].values.max()
    _x = 0
    xticks = []
    for chrom_i in sp.unique(df["chrom"].values):
        _df = df[df["chrom"] == chrom_i]
        if chrom_i % 2 == 0:
            color = colors[0]
        else:
            color = colors[1]
        ax.plot(_df["pos"] + _x, -sp.log10(_df["pv"]), ".", color=color)
        if callback is not None:
            callback(_df)
        xticks.append(_x + 0.5 * _df["pos"].values.max())
        _x += _df["pos"].values.max() + dx
    ax.set_xticks(xticks)
    ax.set_xticklabels(sp.unique(df["chrom"].values))



import scipy as sp
import pandas as pd


def calc_Ai_beta_s2(yKiy, FKiF, FKiy, df):
    import scipy as sp
    import scipy.linalg as la

    Ai = la.pinv(FKiF)
    beta = sp.dot(Ai, FKiy)
    s2 = (yKiy - sp.dot(FKiy[:, 0], beta[:, 0])) / df
    return Ai, beta, s2


def hatodot(A, B):
    """ should be implemented in C """
    import scipy as sp

    A1 = sp.kron(A, sp.ones((1, B.shape[1])))
    B1 = sp.kron(sp.ones((1, A.shape[1])), B)
    return A1 * B1


def christof_trick(A1, F1, D, A2=None, F2=None):
    if A2 is None:
        A2 = A1
    if F2 is None:
        F2 = F1
    out = sp.zeros([A1.shape[1] * F1.shape[1], A2.shape[1] * F2.shape[1]])
    for c in range(A1.shape[0]):
        Oc = sp.dot(A1[[c], :].T, A2[[c], :])
        Or = sp.dot(F1.T, D[:, [c]] * F2)
        out += sp.kron(Oc, Or)
    return out


def append_res(_bim, res):
    """ add new columns to a pandas dataframe with the specified prefix """
    RV = [_bim]
    for _prefix, _res in res.items():
        _res.columns = ["%s_%s" % (_prefix, _key) for _key in _res.keys()]
        RV.append(_res)
    RV = pd.concat(RV, 1)
    return RV



from time import time

import scipy as sp
import scipy.stats as st

from .lmm_core import LMMCore
from .util import calc_Ai_beta_s2
from .mtlmm_helper import define_helper


class MTLMM(LMMCore):
    r"""
    Multi-trait LMM for single-variant association testing.

    The model is

    .. math::
        \text{vec}\left(\mathbf{Y}\right)\sim\mathcal{N}(
        \underbrace{(\mathbf{A}\otimes\mathbf{F})\text{vec}\left(\mathbf{B}\right)}_{\text{covariates}}+
        \underbrace{(\mathbf{A}_{\text{snp}\otimes\mathbf{g})\mathbf{b}}_{\text{genetics}}+
        \sigma^2\underbrace{
        \mathbf{C}^{\text{(g)}}_0\otimes\mathbf{R} +
        \mathbf{C}^{\text{(n)}}_0\otimes\mathbf{I}
        }_{\text{covariance}})

    where :math:`\otimes` is Kronecker product.

    Importantly, :math:`\mathbf{K}_{\boldsymbol{\theta}_0}`
    is not re-estimated under the laternative model,
    i.e. only :math:`\sigma^2` is learnt.
    The covariance parameters :math:`\boldsymbol{\theta}_0`
    need to be learnt using another module, e.g. limix-core.
    Note that as a consequence of this,
    such an implementation of single-variant analysis does not require the
    specification of the whole covariance but just of a method that
    specifies the product of its inverse by a vector.

    The test :math:`\boldsymbol{\beta}\neq{0}` is done
    in bocks of ``step`` variants,
    where ``step`` can be specifed by the user.

    Parameters
    ----------
    Y : (`N`, `P`) ndarray
        phenotype matrix
    F : (`N`, `L`) ndarray
        row fixed-effect design for covariates.
        (default is an intercept)
    A : (`P`, `W`) ndarray
        col fixed-effect design for covariates (default is eye)
    Asnp : (`P`, `Wg`) ndarray
        col fixed-effect design for genetic variants (default is eye)
    Ki_dot : function
        method that takes an array and returns the dot product of
        the inverse of the covariance and the input array.
    """

    def __init__(self, Y, F=None, A=None, Asnp=None, covar=None):

        if F is None:
            F = sp.ones((Y.shape[0], 1))
        if A is None:
            A = sp.eye(Y.shape[1])
        if Asnp is None:
            Asnp = sp.eye(Y.shape[1])

        # the helper implements methods that are specific to the covariance
        self.helper = define_helper(Y, F, A, Asnp, covar)

        # store useful stuff
        self.Y = Y
        self.F = F
        self.A = A
        self.Asnp = Asnp
        self.df = Y.size - A.shape[1] * F.shape[1]

        # fit null
        self._fit_null()

    def _fit_null(self):
        """ Internal functon. Fits the null model """
        self.WKiy = self.helper.get_WKiy()
        self.yKiy = self.helper.get_yKiy()
        self.WKiW = self.helper.get_WKiW()
        # calc beta_F0 and s20
        self.A0i, self.beta_F0, self.s20 = calc_Ai_beta_s2(
            self.yKiy, self.WKiW, self.WKiy, self.df
        )

    def process(self, G, verbose=False):
        r"""
        Fit genotypes one-by-one.

        Parameters
        ----------
        G : (`N`, `S`) ndarray
        verbose : bool
            verbose flag.
        """
        t0 = time()
        k = self.A.shape[1] * self.F.shape[1]
        m = self.Asnp.shape[1]
        W1KiW1 = sp.zeros((k + m, k + m))
        W1KiW1[:k, :k] = self.WKiW
        W1Kiy = sp.zeros((k + m, 1))
        W1Kiy[:k, 0] = self.WKiy[:, 0]
        s2 = sp.zeros(G.shape[1])
        self.beta_g = sp.zeros([m, G.shape[1]])
        for s in range(G.shape[1]):
            self.helper.set_g(G[:, [s]])
            W1Kiy[k:, 0] = self.helper.get_W1Kiy_2()[:, 0]
            W1KiW1[:k, k:] = self.helper.get_W1KiW1_12()
            W1KiW1[k:, :k] = W1KiW1[:k, k:].T
            W1KiW1[k:, k:] = self.helper.get_W1KiW1_22()
            # the following can be sped up by using block matrix inversion, etc
            _, beta, s2[s] = calc_Ai_beta_s2(self.yKiy, W1KiW1, W1Kiy, self.df)
            self.beta_g[:, s] = beta[k:, 0]
        # dlml and pvs
        self.lrt = -self.df * sp.log(s2 / self.s20)
        self.pv = st.chi2(m).sf(self.lrt)

        t1 = time()
        if verbose:
            print("Tested for %d variants in %.2f s" % (G.shape[1], t1 - t0))

        return self.pv, self.beta_g
